﻿package com.dango_itimi.graphics.codec {
	
	import flash.utils.ByteArray;

	public class PNGEncoder {

	    private var crcTable:Vector.<uint>;
		private var encodeType:EncodeType;
		private var imageWidth:int;
		private var imageHeight:int;
		private var main:Function;
		private var png:ByteArray;
		private var IDAT:ByteArray;
		private var FRAME_ENCODE_PIXEL_LIMIT:uint;
		private var countY:uint;
		private var countX:uint;
		private var progressCount:int;
		
		private var PROGRESS_TOTAL:uint;

		public function PNGEncoder(
			encodeType:EncodeType, 
			imageWidth:int, 
			imageHeight:int,
			FRAME_ENCODE_PIXEL_LIMIT:uint = 0
		) {
			this.FRAME_ENCODE_PIXEL_LIMIT = FRAME_ENCODE_PIXEL_LIMIT;
			this.imageHeight = imageHeight;
			this.imageWidth = imageWidth;
			this.encodeType = encodeType;
			
			var encodedPixelTotal:uint = imageWidth * imageHeight;
			
			//buildIDATAsync メソッド終了後実行する 分割メソッド数分数値を追加
			PROGRESS_TOTAL = encodedPixelTotal + 1;
			
			countX = 0;
			countY = 0;
			progressCount = 0;
			
			main = (FRAME_ENCODE_PIXEL_LIMIT == 0) ? batch : initializeCRCTable;
		}
		
		public function run():ByteArray{
			
			main();
			return (main == finished) ? png : null;
		}
		private function batch():void{
			
			initializeCRCTable();
			initializePng();
			buildIDATBatch();
			finishEncode();
		}
		private function initializeCRCTable():void {
			
	        crcTable = new Vector.<uint>();
	
	        for (var n:uint = 0; n < 256; n++) {
	        	
	            var c:uint = n;
	            
	            for (var k:uint = 0; k < 8; k++) 
	                c = (c & 1) ? uint(uint(0xedb88320) ^ uint(c >>> 1)) : uint(c >>> 1);
	            
	            crcTable[n] = c;
	        }
	        main = initializePng;
		}
		private function initializePng():void{
				
	        png = new ByteArray();
	        png.writeUnsignedInt(0x89504E47);
	        png.writeUnsignedInt(0x0D0A1A0A);
	
	        // Build IHDR chunk
	        var IHDR:ByteArray = new ByteArray();
	        IHDR.writeInt(imageWidth);
	        IHDR.writeInt(imageHeight);
			IHDR.writeByte(8); // bit depth per channel
			IHDR.writeByte(6); // color type: RGBA
			IHDR.writeByte(0); // compression method
			IHDR.writeByte(0); // filter method
	        IHDR.writeByte(0); // interlace method
	        writeChunk(png, 0x49484452, IHDR);
	        
	        IDAT = new ByteArray();
	    	encodeType.initialized(IDAT);
	    	
	        main = buildIDAT;
		}
		private function buildIDATBatch():void{
			
	        for (var y:int = 0; y < imageHeight; y++) {
	        	
	            IDAT.writeByte(0);
                for (var x:int = 0; x < imageWidth; x++) encodeType.buildIDAT(x, y);
	        }
		}
		private function buildIDAT():void{
			
			for (var i:Number = 0; i < FRAME_ENCODE_PIXEL_LIMIT; i += imageWidth) {
				
				IDAT.writeByte(0);
	            
	            do encodeType.buildIDAT(countX, countY);
	            while(++countX < imageWidth);
	            
            	countX = 0;
	            progressCount += imageWidth;
				if(++countY < imageHeight) continue;
                
            	main = finishEncode;
            	return;
	        }
		}
		private function finishEncode():void{
	        
	        IDAT.compress();
	        writeChunk(png, 0x49444154, IDAT);
	
	        // Build IEND chunk
	        writeChunk(png, 0x49454E44, null);
	        png.position = 0;
	        
	        progressCount++;
	        main = finished;
	    }
	    private function finished():void{
	    }
	    
	    public function getProgress():Number{
	    	
			return progressCount / PROGRESS_TOTAL * 100;
		}

		/**
		 * 
		 */
		private function writeChunk(png:ByteArray, type:uint, data:ByteArray):void {
			
	        // Write length of data.
	        var len:uint = 0;
	        if(data) len = data.length;
			png.writeUnsignedInt(len);
	        
			// Write chunk type.
			var typePos:uint = png.position;
			png.writeUnsignedInt(type);
	        
			// Write data.
			if(data) png.writeBytes(data);
	
	        // Write CRC of chunk type and data.
			var crcPos:uint = png.position;
	        png.position = typePos;
	        
	        var crc:uint = 0xFFFFFFFF;
	        for (var i:uint = typePos; i < crcPos; i++)
	            crc = uint(crcTable[(crc ^ png.readUnsignedByte()) & uint(0xFF)] ^ uint(crc >>> 8));
	        
	        crc = uint(crc ^ uint(0xFFFFFFFF));
	        png.position = crcPos;
	        png.writeUnsignedInt(crc);
		}
	}
}
